"""Helper functions"""


def strtobool(value: str) -> bool:
    """Checks if string is truthy"""
    lowered = value.lower()
    if lowered in ("y", "yes", "on", "1", "true", "t"):
        return True
    return False



"""Data models"""

from dataclasses import dataclass

from .helpers import strtobool


@dataclass
class KatObligation:
    """Single obligation model."""

    unit_group: int
    status: int
    amount: int
    discount_amount: int
    discount_percentage: int
    description: str
    is_served: bool | None
    vehicle_number: str
    date_breach: str
    date_issued: str
    document_series: str
    document_number: str
    breach_of_order: str

    def __init__(self, unit_group: int, obligation: any):
        """Parse the data."""

        self.unit_group = unit_group
        self.status = obligation["status"]
        self.amount = obligation["amount"]
        self.discount_amount = obligation["discountAmount"]
        self.discount_percentage = int(
            obligation["additionalData"]["discount"])
        self.description = obligation["paymentReason"]

        if "isServed" in obligation["additionalData"]:
            self.is_served = strtobool(
                obligation["additionalData"]["isServed"])
        else:
            self.is_served = False

        self.vehicle_number = obligation["additionalData"]["vehicleNumber"]
        self.date_breach = obligation["additionalData"]["breachDate"]
        self.date_issued = obligation["additionalData"]["issueDate"]
        self.document_series = obligation["additionalData"]["documentSeries"]
        self.document_number = obligation["additionalData"]["documentNumber"]
        self.breach_of_order = obligation["additionalData"]["breachOfOrder"]


@dataclass
class KatObligationUnitGroup:
    """Obligation unit group entry."""

    unit_group: int
    error_no_data_found: bool
    error_reading_data: bool
    obligations: list[KatObligation]

    def __init__(self, unitgroup: any):
        """Parse the data."""

        self.unit_group = unitgroup["unitGroup"]
        self.error_no_data_found = unitgroup["errorNoDataFound"]
        self.error_reading_data = unitgroup["errorReadingData"]

        self.obligations = []
        for ob in unitgroup["obligations"]:
            self.obligations.append(KatObligation(self.unit_group, ob))


@dataclass
class KatObligationApiResponse:
    """Full KAT API Response"""

    obligations_data: list[KatObligationUnitGroup]

    def __init__(self, data: any):
        """Parse the data."""

        self.obligations_data = []
        for od in data["obligationsData"]:
            self.obligations_data.append(KatObligationUnitGroup(od))



"""Error definitions"""

from enum import Enum


class KatErrorType(Enum):
    """Different KAT api error types"""

    VALIDATION_EGN_INVALID = 1
    VALIDATION_ID_DOCUMENT_INVALID = 2
    VALIDATION_USER_NOT_FOUND_ONLINE = 3
    API_TOO_MANY_REQUESTS = 5
    API_ERROR_READING_DATA = 6
    API_UNKNOWN_ERROR = 7
    API_TIMEOUT = 8
    API_INVALID_SCHEMA = 9


class KatError(Exception):
    """Error wrapper"""

    error_type: KatErrorType
    error_message: str

    def __init__(self, error_type: KatErrorType, error_message: str, *args: object) -> None:
        super().__init__(*args)
        self.error_type = error_type
        self.error_message = error_message

    def __str__(self):
        return self.error_message



"""Obligations module"""

from json import JSONDecodeError
import re
import httpx
from httpx import AsyncClient

from .errors import KatError, KatErrorType
from .data_models import KatObligationApiResponse, KatObligation

_REQUEST_TIMEOUT = 10
_KAT_INDIVIDUAL_URL = "https://e-uslugi.mvr.bg/api/Obligations/AND?obligatedPersonType=1&additinalDataForObligatedPersonType=1&mode=1&obligedPersonIdent={egn}&drivingLicenceNumber={license_number}"
_КАТ_BUSINESS_URL = "https://e-uslugi.mvr.bg/api/Obligations/AND?obligatedPersonType=2&additinalDataForObligatedPersonType=1&mode=1&obligedPersonIdent={egn}&personalDocumentNumber={govt_id_number}&uic={bulstat}"

ERR_INVALID_EGN = "EGN is not valid."
ERR_INVALID_LICENSE = "Driving License Number is not valid."
ERR_INVALID_USER_DATA = "User data (EGN and Driving license number combination) is not valid."

ERR_INVALID_GOV_ID = "Government ID Number is not valid."
ERR_INVALID_BULSTAT = "BULSTAT is not valid."

ERR_API_TOO_MANY_REQUESTS = "KAT API too many requests for ID={identifier}"
ERR_API_TIMEOUT = "KAT API request timed out for ID={identifier}"
ERR_API_DOWN = "KAT API was unable to process the request. Try again later."
ERR_API_MALFORMED_RESP = " KAT API returned a malformed response: {data}"
ERR_API_UNKNOWN = "KAT API returned an unknown error: {error}"

REGEX_EGN = r"^[0-9]{2}[0,1,2,4][0-9][0-9]{2}[0-9]{4}$"
REGEX_DRIVING_LICENSE = r"^[0-9]{9}$"

# ID Format Supports "123456789" and "AA1234567"
REGEX_GOVT_ID = r"^[0-9]{9}|[A-Z]{2}[0-9]{7}$"
REGEX_BULSTAT = r"^[0-9]{9}$"


class KatApiClient:
    """KAT API manager"""

    def __self__(self):
        """Initialize API client."""

    def __validate_response(self, data: KatObligationApiResponse):
        """Validate if the user is valid"""

        for od in data.obligations_data:
            if od.error_no_data_found is True:
                raise KatError(
                    KatErrorType.VALIDATION_USER_NOT_FOUND_ONLINE, ERR_INVALID_USER_DATA)

            if od.error_reading_data is True:
                raise KatError(
                    KatErrorType.API_ERROR_READING_DATA, ERR_API_DOWN)

    async def _get_obligations_from_url(
        self, url: str, identifier: str, external_httpx_client: AsyncClient = None
    ) -> list[KatObligation]:
        """
        Gets a list of obligations/fines from URL

        :param url: URL to fetch the data from
        :param identifier: Person identifier - Government ID Number or Driving License Number

        """
        data = {}

        try:
            if external_httpx_client:
                resp = await external_httpx_client.get(url, timeout=_REQUEST_TIMEOUT)
                data = resp.json()
                resp.raise_for_status()
            else:
                async with httpx.AsyncClient() as client:
                    resp = await client.get(url, timeout=_REQUEST_TIMEOUT)

                    resp.raise_for_status()

                    if (resp.headers.get("content-type") == "text/html" and "Достигнат е максимално допустимият брой заявки към системата" in resp.text):
                        raise KatError(
                            KatErrorType.API_TOO_MANY_REQUESTS,
                            ERR_API_TOO_MANY_REQUESTS.format(
                                identifier=identifier)
                        )

                    data = resp.json()

        except httpx.TimeoutException as ex_timeout:
            raise KatError(KatErrorType.API_TIMEOUT, ERR_API_TIMEOUT.format(
                identifier=identifier)) from ex_timeout

        except httpx.HTTPError as ex_apierror:
            raise KatError(KatErrorType.API_UNKNOWN_ERROR, ERR_API_UNKNOWN.format(
                error=str(ex_apierror))) from ex_apierror

        except JSONDecodeError as ex_decode_err:
            raise KatError(KatErrorType.API_UNKNOWN_ERROR, ERR_API_MALFORMED_RESP.format(
                data=str(ex_decode_err))) from ex_decode_err

        if "obligationsData" not in data:
            # This should never happen.
            # If we go in this if, this probably means they changed their schema
            raise KatError(KatErrorType.API_INVALID_SCHEMA,
                           ERR_API_MALFORMED_RESP.format(data=data))

        api_data = KatObligationApiResponse(data)
        self.__validate_response(api_data)

        response = []
        for og in api_data.obligations_data:
            for ob in og.obligations:
                response.append(ob)

        return response

    async def validate_credentials_individual(
            self,
            egn: str,
            license_number: str,
            external_httpx_client: AsyncClient = None) -> bool:
        """
        Validates the combination of EGN and License number for an individual

        :param person_egn: EGN (National Identification Number)
        :param driving_license_number: Driver's License Number

        """

        # Validate EGN
        if egn is None or re.search(REGEX_EGN, egn) is None:
            raise KatError(KatErrorType.VALIDATION_EGN_INVALID,
                           ERR_INVALID_EGN)

        # Validate License Number
        if license_number is None or re.search(REGEX_DRIVING_LICENSE, license_number) is None:
            raise KatError(
                KatErrorType.VALIDATION_ID_DOCUMENT_INVALID, ERR_INVALID_LICENSE)

        data = await self.get_obligations_individual(egn, license_number, external_httpx_client)

        return data is not None

    async def validate_credentials_business(
            self,
            egn: str,
            govt_id_number: str,
            bulstat: str,
            external_httpx_client: AsyncClient = None) -> bool:
        """
        Validates the combination of EGN, Government ID Number and BULSTAT for a business

        :param person_egn: EGN (National Identification Number)
        :param govt_id_number: Government ID Number
        :param bulstat: Business BULSTAT

        """

        # Validate EGN
        if egn is None or re.search(REGEX_EGN, egn) is None:
            raise KatError(KatErrorType.VALIDATION_EGN_INVALID,
                           ERR_INVALID_EGN)

        # Validate Government ID Number
        if govt_id_number is None or re.search(REGEX_GOVT_ID, govt_id_number) is None:
            raise KatError(
                KatErrorType.VALIDATION_ID_DOCUMENT_INVALID, ERR_INVALID_GOV_ID)

        # Validate BULSTAT
        if govt_id_number is None or re.search(REGEX_BULSTAT, bulstat) is None:
            raise KatError(
                KatErrorType.VALIDATION_ID_DOCUMENT_INVALID, ERR_INVALID_BULSTAT)

        data = await self.get_obligations_business(egn, govt_id_number, bulstat, external_httpx_client)

        return data is not None

    async def get_obligations_individual(
        self, egn: str, license_number: str, external_httpx_client: AsyncClient = None
    ) -> list[KatObligation]:
        """
        Gets a list of obligations/fines for an individual

        :param person_egn: EGN (National Identification Number)
        :param driving_license_number: Driver's License Number
        """

        url = _KAT_INDIVIDUAL_URL.format(
            egn=egn, license_number=license_number)

        return await self._get_obligations_from_url(url, license_number, external_httpx_client)

    async def get_obligations_business(
        self, egn: str, govt_id_number: str, bulstat: str, external_httpx_client: AsyncClient = None
    ) -> list[KatObligation]:
        """
        Gets a list of obligations/fines for an individual

        :param person_egn: EGN (National Identification Number)
        :param driving_license_number: Driver's License Number
        """

        url = _КАТ_BUSINESS_URL.format(
            egn=egn, govt_id_number=govt_id_number, bulstat=bulstat)

        return await self._get_obligations_from_url(url, govt_id_number, external_httpx_client)



"""Obligations tests."""

import pytest

import httpx
from pytest_httpx import HTTPXMock

from kat_bulgaria.kat_api_client import (
    KatApiClient,
    KatError,
    KatErrorType,
    ERR_INVALID_EGN,
    ERR_INVALID_LICENSE,
    ERR_INVALID_USER_DATA,
    ERR_API_DOWN
)

from .conftest import EGN, LICENSE, INVALID_EGN, INVALID_LICENSE


# region verify_credentials


@pytest.mark.asyncio
async def test_verify_credentials_success(
    httpx_mock: HTTPXMock, ok_no_fines: pytest.fixture
) -> None:
    """Verify credentials - success."""

    httpx_mock.add_response(json=ok_no_fines)

    resp = await KatApiClient().validate_credentials_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert resp is True


@pytest.mark.asyncio
async def test_verify_credentials_local_invalid_egn(httpx_mock: HTTPXMock) -> None:
    """Verify credentials - local EGN validation failed."""

    with pytest.raises(KatError) as ctx:
        await KatApiClient().validate_credentials_individual(INVALID_EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 0
    assert isinstance(ctx.value, KatError)
    assert ctx.value.error_type == KatErrorType.VALIDATION_EGN_INVALID
    assert ctx.value.error_message == ERR_INVALID_EGN


@pytest.mark.asyncio
async def test_verify_credentials_local_invalid_driver_license(httpx_mock: HTTPXMock) -> None:
    """Verify credentials - local Driver License validation failed."""

    with pytest.raises(KatError) as ctx:
        await KatApiClient().validate_credentials_individual(EGN, INVALID_LICENSE)

    assert len(httpx_mock.get_requests()) == 0
    assert isinstance(ctx.value, KatError)
    assert ctx.value.error_type == KatErrorType.VALIDATION_ID_DOCUMENT_INVALID
    assert ctx.value.error_message == ERR_INVALID_LICENSE


@pytest.mark.asyncio
async def test_verify_credentials_api_invalid_user_data_sent(
    httpx_mock: HTTPXMock, err_nodatafound: pytest.fixture
) -> None:
    """Verify credentials - no user found for credentials"""

    httpx_mock.add_response(json=err_nodatafound, status_code=200)

    with pytest.raises(KatError) as ctx:
        await KatApiClient().validate_credentials_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert isinstance(ctx.value, KatError)
    assert ctx.value.error_type == KatErrorType.VALIDATION_USER_NOT_FOUND_ONLINE
    assert ctx.value.error_message == ERR_INVALID_USER_DATA


@pytest.mark.asyncio
async def test_verify_credentials_api_timeout(httpx_mock: HTTPXMock) -> None:
    """Verify credentials - remote KAT API timeout."""

    httpx_mock.add_exception(httpx.TimeoutException(""))

    with pytest.raises(KatError) as ctx:
        await KatApiClient().validate_credentials_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert isinstance(ctx.value, KatError)
    assert ctx.value.error_type == KatErrorType.API_TIMEOUT
    assert "request timed out for" in ctx.value.error_message


@pytest.mark.asyncio
async def test_verify_credentials_api_down(
    httpx_mock: HTTPXMock, err_apidown: pytest.fixture
) -> None:
    """Verify credentials - remote KAT API returns reading error."""

    httpx_mock.add_response(json=err_apidown, status_code=200)

    with pytest.raises(KatError) as ctx:
        await KatApiClient().validate_credentials_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert isinstance(ctx.value, KatError)
    assert ctx.value.error_type == KatErrorType.API_ERROR_READING_DATA
    assert ctx.value.error_message == ERR_API_DOWN


@pytest.mark.asyncio
async def test_verify_credentials_non_success_status_code(
    httpx_mock: HTTPXMock, ok_no_fines: pytest.fixture
) -> None:
    """Verify credentials - remote KAT API returns error."""

    httpx_mock.add_response(json=ok_no_fines, status_code=400)

    with pytest.raises(KatError) as ctx:
        await KatApiClient().validate_credentials_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert isinstance(ctx.value, KatError)
    assert ctx.value.error_type == KatErrorType.API_UNKNOWN_ERROR
    assert "unknown error" in ctx.value.error_message


@pytest.mark.asyncio
async def test_verify_credentials_api_html_returned(
    httpx_mock: HTTPXMock, err_random_html: pytest.fixture
) -> None:
    """Check obligations - html returned."""

    httpx_mock.add_response(status_code=200, html=err_random_html, headers={
                            'content-type': 'text/html'})

    with pytest.raises(KatError) as ctx:
        await KatApiClient().validate_credentials_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert isinstance(ctx.value, KatError)
    assert ctx.value.error_type == KatErrorType.API_UNKNOWN_ERROR
    assert "malformed response" in ctx.value.error_message


@pytest.mark.asyncio
async def test_verify_credentials_api_too_many_requests(
    httpx_mock: HTTPXMock, err_too_many_requests: pytest.fixture
) -> None:
    """Check obligations - too many requests."""

    httpx_mock.add_response(status_code=200, html=err_too_many_requests, headers={
                            'content-type': 'text/html'})

    with pytest.raises(KatError) as ctx:
        await KatApiClient().validate_credentials_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert isinstance(ctx.value, KatError)
    assert ctx.value.error_type == KatErrorType.API_TOO_MANY_REQUESTS
    assert "too many requests" in ctx.value.error_message

# endregion


# region check_obligations


@pytest.mark.asyncio
async def test_check_obligations_no_fines(
    httpx_mock: HTTPXMock, ok_no_fines: pytest.fixture
) -> None:
    """Check obligations - None."""

    httpx_mock.add_response(json=ok_no_fines)

    resp = await KatApiClient().get_obligations_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert len(resp) == 0


@pytest.mark.asyncio
async def test_check_obligations_field_mapping_success(
    httpx_mock: HTTPXMock, ok_fine_served: pytest.fixture
) -> None:
    """Check obligations - verify field mappings is successful."""

    httpx_mock.add_response(json=ok_fine_served)

    resp = await KatApiClient().get_obligations_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert len(resp) == 1

    obligation = resp[0]

    assert obligation.unit_group == 1
    assert obligation.status == 0
    assert obligation.amount == 100
    assert obligation.discount_amount == 70
    assert obligation.discount_percentage == 30
    assert obligation.description == "ЕЛ.ФИШ СЕРИЯ K 0000000 29.03.2024"
    assert obligation.is_served is True
    assert obligation.vehicle_number == "PB 0000 АА"
    assert obligation.date_breach == "2024-01-25"
    assert obligation.date_issued == "2024-03-29"
    assert obligation.document_series == "K"
    assert obligation.document_number == "123456"
    assert obligation.breach_of_order == "чл. 21, ал. 1, от ЗДвП"


@pytest.mark.asyncio
async def test_check_obligations_has_served(
    httpx_mock: HTTPXMock, ok_fine_served: pytest.fixture
) -> None:
    """Check obligations - has served."""

    httpx_mock.add_response(json=ok_fine_served)

    resp = await KatApiClient().get_obligations_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert len(resp) == 1
    assert resp[0].is_served is True


@pytest.mark.asyncio
async def test_check_obligations_has_not_served(
    httpx_mock: HTTPXMock, ok_fine_not_served: pytest.fixture
) -> None:
    """Check obligations - has NOT served."""

    httpx_mock.add_response(json=ok_fine_not_served)

    resp = await KatApiClient().get_obligations_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert len(resp) == 1
    assert resp[0].is_served is False


@pytest.mark.asyncio
async def test_check_obligations_invalid_user_data_sent(
        httpx_mock: HTTPXMock, err_nodatafound: pytest.fixture
) -> None:
    """Check obligations - local EGN validation failed."""

    httpx_mock.add_response(json=err_nodatafound)

    with pytest.raises(KatError) as ctx:
        await KatApiClient().get_obligations_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert isinstance(ctx.value, KatError)
    assert ctx.value.error_type == KatErrorType.VALIDATION_USER_NOT_FOUND_ONLINE
    assert ctx.value.error_message == ERR_INVALID_USER_DATA


@pytest.mark.asyncio
async def test_check_obligations_api_timeout(httpx_mock: HTTPXMock) -> None:
    """Check obligations - remote KAT API timeout."""

    httpx_mock.add_exception(httpx.TimeoutException(""))

    with pytest.raises(KatError) as ctx:
        await KatApiClient().get_obligations_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert isinstance(ctx.value, KatError)
    assert ctx.value.error_type == KatErrorType.API_TIMEOUT
    assert "request timed out for" in ctx.value.error_message


@pytest.mark.asyncio
async def test_check_obligations_api_down(
    httpx_mock: HTTPXMock, err_apidown: pytest.fixture
) -> None:
    """Check obligations - remote KAT API returns reading error."""

    httpx_mock.add_response(json=err_apidown, status_code=200)

    with pytest.raises(KatError) as ctx:
        await KatApiClient().get_obligations_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert isinstance(ctx.value, KatError)
    assert ctx.value.error_type == KatErrorType.API_ERROR_READING_DATA
    assert "unable to process the request" in ctx.value.error_message


@pytest.mark.asyncio
async def test_check_obligations_non_success_status_code(
    httpx_mock: HTTPXMock, ok_no_fines: pytest.fixture
) -> None:
    """Check obligations - remote KAT API returns error."""

    httpx_mock.add_response(json=ok_no_fines, status_code=400)

    with pytest.raises(KatError) as ctx:
        await KatApiClient().get_obligations_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert isinstance(ctx.value, KatError)
    assert ctx.value.error_type == KatErrorType.API_UNKNOWN_ERROR
    assert "unknown error" in ctx.value.error_message


@pytest.mark.asyncio
async def test_check_obligations_sample2(
    httpx_mock: HTTPXMock, ok_sample2_6fines: pytest.fixture
) -> None:
    """Check obligations - has served."""

    httpx_mock.add_response(json=ok_sample2_6fines)

    resp = await KatApiClient().get_obligations_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert len(resp) == 6
    assert sum(o.is_served for o in resp) == 2
    assert sum(o.amount for o in resp) == 600


@pytest.mark.asyncio
async def test_check_obligations_api_html_returned(
    httpx_mock: HTTPXMock, err_random_html: pytest.fixture
) -> None:
    """Check obligations - html returned."""

    httpx_mock.add_response(status_code=200, html=err_random_html, headers={
                            'content-type': 'text/html'})

    with pytest.raises(KatError) as ctx:
        await KatApiClient().get_obligations_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert isinstance(ctx.value, KatError)
    assert ctx.value.error_type == KatErrorType.API_UNKNOWN_ERROR
    assert "malformed response" in ctx.value.error_message


@pytest.mark.asyncio
async def test_check_obligations_api_too_many_requests(
    httpx_mock: HTTPXMock, err_too_many_requests: pytest.fixture
) -> None:
    """Check obligations - too many requests."""

    httpx_mock.add_response(status_code=200, html=err_too_many_requests, headers={
                            'content-type': 'text/html'})

    with pytest.raises(KatError) as ctx:
        await KatApiClient().get_obligations_individual(EGN, LICENSE)

    assert len(httpx_mock.get_requests()) == 1
    assert isinstance(ctx.value, KatError)
    assert ctx.value.error_type == KatErrorType.API_TOO_MANY_REQUESTS
    assert "too many requests" in ctx.value.error_message

# endregion
